@@ -25,6 +25,7 @@
 #include <ctype.h>
 #include <errno.h>
 #include <stdio.h>
+#include <signal.h>
 #include <unistd.h>
 #include <getopt.h>
 #include <shadow.h>
@@ -32,16 +33,20 @@
 #include <libintl.h>
 #include <security/pam_appl.h>
 #include <security/pam_misc.h>
+#ifdef WITH_SELINUX
+#include <selinux/selinux.h>
+#include <selinux/av_permissions.h>
+#endif
 
-#include "getdef.h"
 #include "public.h"
-#include "main.h"
+#include "logindefs.h"
+#include "error_codes.h"
 
 #ifndef _
 #define _(String) gettext (String)
 #endif
 
-static int shadow_chfn = 1;
+static int shadow_chfn = 0;
 
 struct fn_info {
   char *fullname;
@@ -117,28 +122,17 @@ may_change_field (int field)
 {
   const char *cp;
 
-  /* CHFN_RESTRICT can now specify exactly which fields may be changed
-     by regular users, by using any combination of the following
-     letters:
-     f - full name
-     r - room number
-     w - work phone
-     h - home phone
-
-     This makes it possible to disallow changing the room number
-     information, for example - this feature was suggested by Maciej
-     'Tycoon' Majchrowski.
-
-     For backward compatibility, "yes" is equivalent to "rwh",
-     "no" is equivalent to "frwh". Only root can change anything
-     if the string is empty or not defined at all.  */
-
+  /* root is always allowed to change everything.  */
   if (getuid () == 0)
     return 1;
 
-  cp = getdef_str ("CHFN_RESTRICT");
+  cp = getlogindefs_str ("CHFN_RESTRICT", "");
+  /* CHFN_RESTRICT specifies exactly which fields may be changed
+     by regular users.  */
   if (!cp)
     cp = "";
+  /*  For backward compatibility, "yes" is equivalent to "rwh",
+      "no" is equivalent to "frwh".  */
   else if (strcmp (cp, "yes") == 0)
     cp = "rwh";
   else if (strcmp (cp, "no") == 0)
@@ -156,6 +150,9 @@ get_fields (fn_info *old, fn_info *new)
 {
   printf (_("Enter the new value, or press ENTER for the default\n"));
 
+  /* Allow the user to abort with Ctrl-C.  */
+  signal (SIGINT, SIG_DFL);
+
   if (may_change_field ('f'))
     new->fullname = get_value (old->fullname, _("Full Name"));
   else
@@ -180,6 +177,8 @@ get_fields (fn_info *old, fn_info *new)
     new->other = get_value (old->other, _("Other"));
   else
     printf (_("\tOther: %s\n"), old->other ?:"");
+
+  signal (SIGINT, SIG_IGN);
 }
 
 /* parse_passwd () -- take a struct password and fill in the
@@ -233,7 +232,7 @@ check_field (const char *program, const char *field, const char *illegal)
 }
 
 int
-chfn_main (const char *program, int argc, char *argv[])
+main (int argc, char *argv[])
 {
   uid_t uid = getuid ();
   user_t *pw_data = NULL;
@@ -274,12 +273,12 @@ chfn_main (const char *program, int argc, char *argv[])
   struct option *long_options;
   char *optstring;
   char *envstr;
+  const char *program = basename (argv[0]);
 
   setlocale(LC_ALL, "");
   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
 
-  program = basename (argv[0]);
   if (strcasecmp (program, "ypchfn") == 0)
     use_service = "nis";
   else if (strcasecmp (program, "chfn") != 0)
@@ -288,6 +287,12 @@ chfn_main (const char *program, int argc, char *argv[])
       return 1;
     }
 
+  openlog (program, LOG_PID, LOG_AUTHPRIV);
+
+  /* Before going any further, raise the ulimit and ignore
+     signals.  */
+  init_environment ();
+
   envstr = getenv ("SHADOW_CHFN");
   if (envstr)
     {
@@ -424,7 +429,7 @@ chfn_main (const char *program, int argc, char *argv[])
 	    print_help_util (program);
           return 0;
         case 'v':
-          print_version (program);
+          print_version (program, "2003");
           return 0;
         case 'u':
 	  if (shadow_chfn)
@@ -486,7 +491,7 @@ chfn_main (const char *program, int argc, char *argv[])
 	    }
 	  if (!pw)
 	    {
-	      fprintf (stderr, _("%s: Unknown user %s\n"), program, user);
+	      fprintf (stderr, _("%s: Unknown user `%s'.\n"), program, user);
 	      return 1;
 	    }
 	}
@@ -495,7 +500,7 @@ chfn_main (const char *program, int argc, char *argv[])
       if (pw_data == NULL || pw_data->service == S_NONE)
 	{
 	  if (use_service)
-	    fprintf (stderr, _("%s: User %s is not known to service %s\n"),
+	    fprintf (stderr, _("%s: User `%s' is not known to service `%s'.\n"),
 		     program, pw->pw_name, use_service);
 	  else
 	    fprintf (stderr, _("%s: Unknown user %s\n"), program,
@@ -503,6 +508,27 @@ chfn_main (const char *program, int argc, char *argv[])
 	  return 1;
 	}
     }
+#ifdef WITH_SELINUX
+  if (is_selinux_enabled ())
+    {
+      if ((uid == 0) &&
+          (selinux_check_access (pw_data->pw.pw_name, PASSWD__CHFN) != 0))
+        {
+          security_context_t user_context;
+          if (getprevcon (&user_context) < 0)
+            user_context =
+              (security_context_t) strdup (_("Unknown user context"));
+          syslog (LOG_ALERT,
+	    "%s is not authorized to change the finger information for %s\n",
+                  user_context, pw_data->pw.pw_name);
+           fprintf (stderr,
+                 _("%s: %s is not authorized to change the finger information for %s\n"),
+                    program, user_context, pw_data->pw.pw_name);
+           freecon (user_context);
+           return E_NOPERM;
+        }
+    }
+#endif
 
   /* Only root is allowed to change the gecos field for local users. */
   if (uid && uid != pw_data->pw.pw_uid && pw_data->service == S_LOCAL)
@@ -518,8 +544,14 @@ chfn_main (const char *program, int argc, char *argv[])
   if (!silent)
     printf (_("Changing finger information for %s.\n"), pw_data->pw.pw_name);
 
-  if (do_authentication (program, caller_name, pw_data, 1) != 0)
-    return 1;
+  if (getlogindefs_bool ("CHFN_AUTH", 1) || pw_data->service != S_LOCAL)
+    {
+      if (do_authentication (program, caller_name, pw_data) != 0)
+	return 1;
+      if (pw_data->service != S_LOCAL && get_old_clear_password (pw_data) != 0)
+	return 1;
+    }
+
   {
     fn_info old;
     int len;
@@ -598,9 +630,9 @@ chfn_main (const char *program, int argc, char *argv[])
 
   pw_data->new_gecos = new_gecos;
 
-  if (do_setpwnam (pw_data) != 0)
+  if (write_user_data (pw_data, 0) != 0)
     {
-      fprintf (stderr, "Error while changing finger information\n");
+      fprintf (stderr, _("Error while changing finger information.\n"));
       free_user_t (pw_data);
       return 1;
     }
