@@ -1,4 +1,4 @@
-/* Copyright (C) 2002, 2003 Thorsten Kukuk
+/* Copyright (C) 2003 Thorsten Kukuk
    Author: Thorsten Kukuk <kukuk@suse.de>
 
    This program is free software; you can redistribute it and/or modify
@@ -34,16 +34,32 @@
 #include <security/pam_misc.h>
 #include <locale.h>
 #include <libintl.h>
+#ifdef WITH_SELINUX
+#include <selinux/selinux.h>
+#include <selinux/av_permissions.h>
+#endif
 
 #ifndef _
 #define _(String) gettext (String)
 #endif
 
-#include "main.h"
-#include "getdef.h"
 #include "public.h"
+#include "logindefs.h"
+#include "read-files.h"
 #include "error_codes.h"
 
+#ifndef GPASSWD_PROGRAM
+#define GPASSWD_PROGRAM "gpasswd"
+#endif
+
+#ifndef CHFN_PROGRAM
+#define CHFN_PROGRAM "chfn"
+#endif
+
+#ifndef CHSH_PROGRAM
+#define CHSH_PROGRAM "chsh"
+#endif
+
 /* How often should we try to lock the passwd database ?  */
 #define MAX_LOCK_RETRIES 3
 
@@ -109,38 +125,6 @@ display_pw (const struct passwd *pw)
     printf ("%s %s\n", pw->pw_name, pw_status (pw->pw_passwd));
 }
 
-/*
- * passwd - change a user's password file information
- *
- *      This command controls the password file and commands which are
- *      used to modify it.
- *
- *      The valid options are
- *
- *      -l      lock the named account (*)
- *      -u      unlock the named account (*)
- *      -d      delete the password for the named account (*)
- *      -e      expire the password for the named account (*)
- *      -x #    set sp_max to # days (*)
- *      -n #    set sp_min to # days (*)
- *      -w #    set sp_warn to # days (*)
- *      -i #    set sp_inact to # days (*)
- *      -h      change home directory (*)
- *      -S      show password status of named account
- *      -a      with -S show password status for all accounts
- *      -g      execute gpasswd command to interpret flags
- *      -f      execute chfn command to interpret flags
- *      -s      execute chsh command to interpret flags
- *      -k      change password only if expired
- *      -D binddn Use binddn for LDAP binding
- *
- *      (*) requires administrator permission to execute.
- *
- *      All of the time fields are entered in days and converted to the
- *      appropriate internal format.  For finer resolute the chage
- *      command must be used.
- */
-
 static void
 print_usage (FILE *stream, const char *program)
 {
@@ -174,15 +158,21 @@ print_help (const char *program)
   fputs (_("  -r service     Use nameservice 'service'\n"), stdout);
   fputs (_("  -D binddn      Use dn \"binddn\" to bind to the LDAP directory\n"),
 	 stdout);
+  fputs (_("  -P path        Search passwd and shadow file in \"path\"\n"),
+	 stdout);
   fputs (_("  --help         Give this help list\n"), stdout);
   fputs (_("  --usage        Give a short usage message\n"), stdout);
   fputs (_("  --version      Print program version\n"), stdout);
   fputs (_("Valid services for -r are: files, nis, nisplus, ldap\n"), stdout);
 }
 
-int
+static int
 passwd_main (const char *program, int argc, char **argv)
 {
+  int buflen = 256;
+  char *buffer = alloca (buflen);
+  struct passwd resultbuf;
+  struct passwd *pw = NULL;
   int admin_only = 0;
   int silent = 0;
   uid_t uid = getuid ();
@@ -192,7 +182,7 @@ passwd_main (const char *program, int argc, char **argv)
   char *binddn = NULL;
   int k_flag = 0, a_flag = 0, d_flag = 0, e_flag = 0, h_flag = 0,
     i_flag = 0, l_flag = 0, n_flag = 0, u_flag = 0, x_flag = 0,
-    S_flag = 0, w_flag = 0;
+    S_flag = 0, w_flag = 0, P_flag = 0;
   long inact = 0, age_min = 0, age_max = 0, warn = 0;
 
   /* Parse program arguments */
@@ -203,6 +193,7 @@ passwd_main (const char *program, int argc, char **argv)
       static struct option long_options[] =
 	{
 	  {"binddn", required_argument, NULL, 'D'},
+	  {"path", required_argument, NULL, 'P'},
 	  {"service", required_argument, NULL, 'r'},
 	  {"version", no_argument, NULL, '\255'},
 	  {"usage", no_argument, NULL, '\254'},
@@ -210,7 +201,7 @@ passwd_main (const char *program, int argc, char **argv)
 	  {NULL, 0, NULL, '\0'}
 	};
 
-      c = getopt_long (argc, argv, "adD:hlqr:uSekn:x:i:w:", long_options,
+      c = getopt_long (argc, argv, "adD:hlqr:uSekn:x:i:w:P:", long_options,
                        &option_index);
       if (c == EOF)
         break;
@@ -282,7 +273,7 @@ passwd_main (const char *program, int argc, char **argv)
 	    use_service = "ldap";
           else
             {
-              fprintf (stderr, _("Service \"%s\" not supported\n"), optarg);
+              fprintf (stderr, _("Service `%s' not supported\n"), optarg);
               print_usage (stderr, program);
               return E_BAD_ARG;
             }
@@ -314,11 +305,15 @@ passwd_main (const char *program, int argc, char **argv)
 	    }
           admin_only = 1;
           break;
+	case 'P':
+	  P_flag = 1;
+	  files_etc_dir = strdup (optarg);
+	  break;
 	case '\253':
           print_help (program);
           return 0;
         case '\255':
-          print_version (program);
+          print_version (program, "2003");
           return 0;
         case '\254':
           print_usage (stdout, program);
@@ -354,8 +349,6 @@ passwd_main (const char *program, int argc, char **argv)
      you must be root.  */
   if (a_flag)
     {
-      struct passwd *pw;
-
       if (admin_only || !S_flag || (argc != 0))
 	{
 	  print_error (program);
@@ -381,11 +374,6 @@ passwd_main (const char *program, int argc, char **argv)
     }
   else
     {
-      int buflen = 256;
-      char *buffer = alloca (buflen);
-      struct passwd resultbuf;
-      struct passwd *pw;
-
       /* Determine our own user name for PAM authentication.  */
       while (getpwuid_r (uid, &resultbuf, buffer, buflen, &pw) != 0
 	     && errno == ERANGE)
@@ -415,23 +403,55 @@ passwd_main (const char *program, int argc, char **argv)
             }
           if (!pw)
             {
-              fprintf (stderr, _("%s: Unknown user %s\n"), program, argv[0]);
+              fprintf (stderr, _("%s: Unknown user `%s'.\n"),
+		       program, argv[0]);
               return E_NOPERM;
             }
         }
+
       pw_data = do_getpwnam (pw->pw_name, use_service);
       if (pw_data == NULL || pw_data->service == S_NONE)
-        {
-          if (use_service)
-            fprintf (stderr, _("%s: User %s is not known to service %s\n"),
-                     program, pw->pw_name, use_service);
-          else
-            fprintf (stderr, _("%s: Unknown user %s\n"), program,
-                     pw->pw_name);
-          return E_NOPERM;
-        }
+	{
+	  if (use_service)
+	    {
+	      fprintf (stderr,
+		       _("%s: User `%s' is not known to service `%s'\n"),
+		       program, pw->pw_name, use_service);
+	      return E_NOPERM;
+	    }
+	  else if (admin_only)
+	    {
+	      /* Only print error, if we need the pw_data informations
+		 later. Else ignore it and let PAM do it.  */
+	      fprintf (stderr, _("%s: Unknown user `%s'.\n"), program,
+		       pw->pw_name);
+	      return E_NOPERM;
+	    }
+	}
     }
 
+#ifdef WITH_SELINUX
+  if (is_selinux_enabled ())
+    {
+      if ((getuid () == 0) &&
+	  (selinux_check_access (pw->pw_name, PASSWD__PASSWD) != 0))
+	{
+	  security_context_t user_context;
+	  if (getprevcon (&user_context) < 0)
+	    user_context =
+	      (security_context_t) strdup (_("Unknown user context"));
+	  syslog (LOG_ALERT,
+		  _("%s is not authorized to change the password of %s\n"),
+                  user_context, pw->pw_name);
+           fprintf (stderr,
+		 _("%s: %s is not authorized to change the password of %s\n"),
+		    program, user_context, pw->pw_name);
+           freecon (user_context);
+           exit (E_NOPERM);
+	}
+    }
+#endif
+
   /* Check if normal users are allowed to change the data. For NIS+ and
      LDAP, we let the service decide if the user is allowed.  */
   if (uid != 0 && pw_data->service != S_NISPLUS && pw_data->service != S_LDAP)
@@ -439,24 +459,27 @@ passwd_main (const char *program, int argc, char **argv)
       if (admin_only)
 	{
           fprintf(stderr, _("%s: Permission denied\n"), program);
-	  free_user_t (pw_data);
+	  if (pw_data)
+	    free_user_t (pw_data);
 	  return E_NOPERM;
 	}
-      if (pw_data->pw.pw_uid != uid)
+      if (pw->pw_uid != uid)
 	{
-	  fprintf (stderr, _("You cannot change the shadow data for \"%s\".\n"),
-		   pw_data->pw.pw_name);
-	  syslog (LOG_WARNING, "%d cannot change shadow data for \"%s\"",
-		  uid, pw_data->pw.pw_name);
-	  free_user_t (pw_data);
+	  fprintf (stderr, _("You cannot change the shadow data for `%s'.\n"),
+		   pw->pw_name);
+	  syslog (LOG_WARNING, "%d cannot change shadow data for `%s'",
+		  uid, pw->pw_name);
+	  if (pw_data)
+	    free_user_t (pw_data);
 	  return E_NOPERM;
 	}
     }
 
   if (S_flag)
     {
-      display_pw (&pw_data->pw);
-      free_user_t (pw_data);
+      display_pw (pw);
+      if (pw_data)
+	free_user_t (pw_data);
       return E_SUCCESS;
     }
 
@@ -466,22 +489,30 @@ passwd_main (const char *program, int argc, char **argv)
       pam_handle_t *pamh = NULL;
       int flags = 0, ret;
 
+      if (P_flag)
+	{
+	  fprintf (stderr, _("%s: -P flag not supported in this mode!\n"),
+		   program);
+	  return E_USAGE;
+	}
+
       if (!silent)
-	printf (_("Changing password for %s.\n"), pw_data->pw.pw_name);
+	printf (_("Changing password for %s.\n"), pw->pw_name);
 
       if (silent)
 	flags |= PAM_SILENT;
       if (k_flag)
 	flags |= PAM_CHANGE_EXPIRED_AUTHTOK;
 
-      ret = pam_start ("passwd", pw_data->pw.pw_name, &conv, &pamh);
+      ret = pam_start ("passwd", pw->pw_name, &conv, &pamh);
       if (ret != PAM_SUCCESS)
 	{
           fprintf (stderr, _("%s: PAM Failure, aborting: %s\n"),
                    program, pam_strerror (pamh, ret));
           syslog (LOG_ERR, "Couldn't initialize PAM: %s",
                   pam_strerror (pamh, ret));
-          free_user_t (pw_data);
+	  if (pw_data)
+	    free_user_t (pw_data);
 	  return E_PAM_ERROR;
 	}
 
@@ -490,9 +521,10 @@ passwd_main (const char *program, int argc, char **argv)
 	{
           syslog (LOG_ERR, "User %s: %s", caller_name,
                   pam_strerror (pamh, ret));
-          sleep (getdef_num ("FAIL_DELAY", 1));
+          sleep (getlogindefs_num ("FAIL_DELAY", 1));
 	  fprintf (stderr, "%s: %s\n", program, pam_strerror (pamh, ret));
-	  free_user_t (pw_data);
+	  if (pw_data)
+	    free_user_t (pw_data);
 	  return E_PAM_ERROR;
 	}
 
@@ -517,22 +549,26 @@ passwd_main (const char *program, int argc, char **argv)
     {
       if (pw_data->use_shadow)
 	{
-	  if (pw_data->sp.sp_pwdp[0] == '!')
+	  /* If the password is only "!", don't unlock it.  */
+	  if (pw_data->sp.sp_pwdp[0] == '!' &&
+	      strlen (pw_data->sp.sp_pwdp) > 1)
 	    pw_data->newpassword = strdup (&pw_data->sp.sp_pwdp[1]);
 	  else
 	    {
-	      fprintf (stderr, _("Cannot unlock the password for \"%s\"!\n"),
+	      fprintf (stderr, _("Cannot unlock the password for `%s'!\n"),
 		       pw_data->pw.pw_name);
 	      return E_FAILURE;
 	    }
 	}
       else
 	{
-	  if (pw_data->pw.pw_passwd[0] == '!')
+	  /* If the password is only "!", don't unlock it.  */
+	  if (pw_data->pw.pw_passwd[0] == '!' &&
+	      strlen (pw_data->pw.pw_passwd) > 1)
 	    pw_data->newpassword = strdup (&pw_data->pw.pw_passwd[1]);
 	  else
 	    {
-	      fprintf (stderr, _("Cannot unlock the password for \"%s\"!\n"),
+	      fprintf (stderr, _("Cannot unlock the password for `%s'!\n"),
 		       pw_data->pw.pw_name);
 	      return E_FAILURE;
 	    }
@@ -553,7 +589,7 @@ passwd_main (const char *program, int argc, char **argv)
 	    }
 	  else
 	    {
-	      fprintf (stderr, _("Password for \"%s\" is already locked!\n"),
+	      fprintf (stderr, _("Password for `%s' is already locked!\n"),
 		       pw_data->pw.pw_name);
 	      return E_FAILURE;
 	    }
@@ -571,7 +607,7 @@ passwd_main (const char *program, int argc, char **argv)
 	    }
 	  else
 	    {
-	      fprintf (stderr, _("Password for \"%s\" is already locked!\n"),
+	      fprintf (stderr, _("Password for `%s' is already locked!\n"),
 		       pw_data->pw.pw_name);
 	      return E_FAILURE;
 	    }
@@ -591,7 +627,7 @@ passwd_main (const char *program, int argc, char **argv)
   if (x_flag || n_flag || w_flag || i_flag || e_flag || e_flag)
     pw_data->sp_changed = TRUE;
 
-  if (do_setpwnam (pw_data) != 0)
+  if (write_user_data (pw_data, 0) != 0)
     {
       if (pw_data->sp_changed)
 	fprintf (stderr,
@@ -617,3 +653,56 @@ passwd_main (const char *program, int argc, char **argv)
 
   return E_SUCCESS;
 }
+
+int
+main (int argc, char **argv)
+{
+  int retval;
+  char *prog;
+
+  setlocale (LC_ALL, "");
+  bindtextdomain (PACKAGE, LOCALEDIR);
+  textdomain (PACKAGE);
+
+  prog = basename (argv[0]);
+
+  openlog (prog, LOG_PID, LOG_AUTHPRIV);
+
+  /* Before going any further, raise the ulimit and ignore
+     signals.  */
+
+  init_environment ();
+
+  /* easy way to get ride of the first argument.  */
+  if (argc > 1 && argv[1][0] == '-' && strchr ("gfs", argv[1][1]))
+    {
+      char buf[200];
+
+      setuid (getuid ());
+      switch (argv[1][1])
+        {
+        case 'g':
+          argv[1] = GPASSWD_PROGRAM;  /* XXX warning: const */
+          break;
+        case 'f':
+          argv[1] = CHFN_PROGRAM;  /* XXX warning: const */
+          break;
+        case 's':
+          argv[1] = CHSH_PROGRAM;  /* XXX warning: const */
+          break;
+        default:
+          /* If this happens we have a real problem. */
+          abort ();
+        }
+      snprintf (buf, sizeof buf, _("passwd: Cannot execute %s"), argv[1]);
+      execvp(argv[1], &argv[1]);
+      perror(buf);
+      syslog (LOG_ERR, "Cannot execute %s", argv[1]);
+      closelog ();
+      return E_FAILURE;
+    }
+
+  retval = passwd_main (prog, argc, argv);
+  closelog ();
+  return retval;
+}
